
# To make tidy-imports automatically transform imports, create a dictionary
# like this:
#   __canonical_imports__ = {
#      "oldmodule.oldname": "newmodule.newname"
#   }
#
# You can write this in any *.py file in your $PYFLYBY_PATH,
# e.g. ~/.pyflyby/canonical.py.

